#include "DriverEntry.h"


NTSTATUS DriverEntry(PDRIVER_OBJECT pDriverObject, PUNICODE_STRING)
{
	NTSTATUS status;
	ULONG i;


	DBG_PRINT("DriverEntry");
	status = createDevice(pDriverObject);
	if (!NT_SUCCESS(status))
		return status;

	for (i=0; i<=IRP_MJ_MAXIMUM_FUNCTION; i++)
		pDriverObject->MajorFunction[i] = dispatchDefault_NotImplemented;

	pDriverObject->MajorFunction[IRP_MJ_CREATE] = dispatchDefault_Success;
	pDriverObject->MajorFunction[IRP_MJ_CLEANUP] = dispatchDefault_Success;
	pDriverObject->MajorFunction[IRP_MJ_CLOSE] = dispatchDefault_Success;
	pDriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = dispatchDeviceControl;

	pDriverObject->DriverUnload = unloadDevice;

	pDriverObject->DeviceObject->Flags &= ~DO_DEVICE_INITIALIZING;

	return status;
}

NTSTATUS createDevice(PDRIVER_OBJECT pDriverObject)
{
	NTSTATUS status;
	PDEVICE_OBJECT pDeviceObject;
	PDEVICE_EXTENSION pDeviceExtension;
	UNICODE_STRING nameNtDevice;
	UNICODE_STRING nameDosDevice;


	DBG_PRINT("createDevice");

	//̏
	RtlInitUnicodeString(&nameNtDevice, NAME_NT_DEVICE);
	RtlInitUnicodeString(&nameDosDevice, NAME_DOS_DEVICE);

	//deviceObject쐬
	status = IoCreateDevice(pDriverObject, sizeof(DEVICE_EXTENSION), &nameNtDevice, FILE_DEVICE_NETWORK, 0, FALSE, &pDeviceObject);
	if (!NT_SUCCESS(status))
	{
		DBG_PRINT("failure IoCreateDevice");

		return status;
	}

	pDeviceObject->Flags |= DO_BUFFERED_IO;
	pDeviceExtension = PDEVICE_EXTENSION(pDeviceObject->DeviceExtension);
	RtlZeroMemory(pDeviceExtension, sizeof(DEVICE_EXTENSION));

	//applicationQƂ₷悤symboliclink쐬
	status = IoCreateSymbolicLink(&nameDosDevice, &nameNtDevice);
	if (!NT_SUCCESS(status))
	{
		DBG_PRINT("failure IoCreateSymbolicLink: %x", status);
		IoDeleteDevice(pDeviceObject);

		return status;
	}

	return status;
}

VOID unloadDevice(PDRIVER_OBJECT pDriverObject)
{
	UNICODE_STRING nameDosDevice;


	RtlInitUnicodeString(&nameDosDevice, NAME_DOS_DEVICE);
	IoDeleteSymbolicLink(&nameDosDevice);
	IoDeleteDevice(pDriverObject->DeviceObject);
}

NTSTATUS dispatchDefault_NotImplemented(PDEVICE_OBJECT pDeviceObject, PIRP pIrp)
{
	NTSTATUS status;


	status = STATUS_NOT_IMPLEMENTED;
	pIrp->IoStatus.Status = status;
	pIrp->IoStatus.Information = 0;
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);

	return status;
}

NTSTATUS dispatchDefault_Success(PDEVICE_OBJECT pDeviceObject, PIRP pIrp)
{
	NTSTATUS status;


	status = STATUS_SUCCESS;
	pIrp->IoStatus.Status = status;
	pIrp->IoStatus.Information = 0;
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);

	return status;
}

NTSTATUS dispatchDeviceControl(PDEVICE_OBJECT pDeviceObject, PIRP pIrp)
{
	NTSTATUS status;
	PDEVICE_EXTENSION pDeviceExtension;
	PIO_STACK_LOCATION pIrpStack;
	PVOID pDataInOut;
	ULONG lengthDataIn;
	SSend *pSendFromUser;
	SSendInDriver sendInDriver;
	PIO_WORKITEM pWork;


	DBG_PRINT("dispatchDeviceControl");

	pDeviceExtension = PDEVICE_EXTENSION(pDeviceObject->DeviceExtension);
	pIrpStack = IoGetCurrentIrpStackLocation(pIrp);
	pDataInOut = pIrp->AssociatedIrp.SystemBuffer;
	lengthDataIn = pIrpStack->Parameters.DeviceIoControl.InputBufferLength;

	if (MSENDUDP_SEND != pIrpStack->Parameters.DeviceIoControl.IoControlCode)
	{
		DBG_PRINT("MSENDUDP_SEND != IoControlCode");
		status = STATUS_NOT_IMPLEMENTED;
		pIrp->IoStatus.Status = status;
		pIrp->IoStatus.Information = 0;
		IoCompleteRequest(pIrp, IO_NO_INCREMENT);

		return status;
	}

	if (NULL==pDataInOut || lengthDataIn!=sizeof(SSend))
	{
		DBG_PRINT("NULL==pDataInOut || lengthDataIn!=sizeof(SSend)");
		status = STATUS_INVALID_PARAMETER;
		pIrp->IoStatus.Status = status;
		pIrp->IoStatus.Information = 0;
		IoCompleteRequest(pIrp, IO_NO_INCREMENT);

		return status;
	}

	pSendFromUser = (SSend *)pDataInOut;
	sendInDriver.m_pData = IoAllocateMdl(pSendFromUser->m_pData, pSendFromUser->m_lengthData, FALSE, FALSE, NULL);
	if (NULL == pIrp)
	{
		DBG_PRINT("can NOT allocate MDL");
		status = STATUS_INSUFFICIENT_RESOURCES;
		pIrp->IoStatus.Status = status;
		pIrp->IoStatus.Information = 0;
		IoCompleteRequest(pIrp, IO_NO_INCREMENT);

		return status;
	}

	__try
	{
		MmProbeAndLockPages(sendInDriver.m_pData, KernelMode, IoReadAccess);
	}
	__except (EXCEPTION_EXECUTE_HANDLER)
	{
		DBG_PRINT("failure probe and lock MDL");
		IoFreeMdl(sendInDriver.m_pData);
		status = STATUS_INSUFFICIENT_RESOURCES;
		pIrp->IoStatus.Status = status;
		pIrp->IoStatus.Information = 0;
		IoCompleteRequest(pIrp, IO_NO_INCREMENT);

		return status;
	}

	KeInitializeEvent(&sendInDriver.m_event, NotificationEvent, FALSE);
	sendInDriver.m_addressRemote = pSendFromUser->m_addressRemote;
	sendInDriver.m_portRemote = pSendFromUser->m_portRemote;

	pWork = IoAllocateWorkItem(pDeviceObject);
	if (NULL == pWork)
	{
		DBG_PRINT("can NOT allocate workItem");
		IoFreeMdl(sendInDriver.m_pData);
		status = STATUS_INSUFFICIENT_RESOURCES;
		pIrp->IoStatus.Status = status;
		pIrp->IoStatus.Information = 0;
		IoCompleteRequest(pIrp, IO_NO_INCREMENT);

		return status;
	}

	IoQueueWorkItem(pWork, worker, CriticalWorkQueue, &sendInDriver);
	KeWaitForSingleObject(&sendInDriver.m_event, Executive, KernelMode, FALSE, NULL);
	IoFreeWorkItem(pWork);

	status = sendInDriver.m_status;
	pIrp->IoStatus.Status = status;
	pIrp->IoStatus.Information = sendInDriver.m_lengthProcessed;
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);

	return status;
}

VOID worker(PDEVICE_OBJECT pDeviceObject, PVOID pContext)
{
	NTSTATUS status;
	SSendInDriver *pSend;
	HANDLE hUdp;
	PFILE_OBJECT pFileUdp;


	pSend = (SSendInDriver *)pContext;

	status = openTransportAddressUdp(&hUdp, &pFileUdp);
	if (!NT_SUCCESS(status))
	{
		DBG_PRINT("can NOT open transportAddress");
		IoFreeMdl(pSend->m_pData);
		pSend->m_status = status;
		pSend->m_lengthProcessed = 0;
		KeSetEvent(&pSend->m_event, EVENT_INCREMENT, FALSE);

		return;
	}

	status = sendDatagram(pFileUdp, pSend->m_addressRemote, pSend->m_portRemote, pSend->m_pData, &pSend->m_lengthProcessed);
	pSend->m_status = status;
	closeTransportAddressUdp(hUdp, pFileUdp);
	KeSetEvent(&pSend->m_event, EVENT_INCREMENT, FALSE);
}

NTSTATUS sendDatagram(PFILE_OBJECT pFileTransportAddress, ULONG addressRemote, USHORT portRemote, PMDL pMdl, PULONG pLengthProcessed)
{
	NTSTATUS status;
	PIRP pIrp;
	PDEVICE_OBJECT pDeviceUdp;
	KEVENT event;
	IO_STATUS_BLOCK statusIo;
	TDI_CONNECTION_INFORMATION infoConnection;
	TA_IP_ADDRESS addressIp;


	DBG_PRINT("sendDatagram");
	if (NULL != pLengthProcessed)
		*pLengthProcessed = 0;

	//̌
	if (NULL == pFileTransportAddress)
	{
		DBG_PRINT("NULL == pFileTransportAddress");
		IoFreeMdl(pMdl);

		return STATUS_INVALID_PARAMETER;
	}

	//IRP쐬
	pDeviceUdp = IoGetRelatedDeviceObject(pFileTransportAddress);
	KeInitializeEvent(&event, NotificationEvent, FALSE);
	pIrp = TdiBuildInternalDeviceControlIrp(TDI_SEND_DATAGRAM, pDeviceUdp, pFileTransportAddress, &event, &statusIo);
	if (NULL == pIrp)
	{
		DBG_PRINT("NULL == pIrp");
		IoFreeMdl(pMdl);

		return STATUS_INSUFFICIENT_RESOURCES;
	}

	//TDI_CONNECTION_INFORMATION̐ݒ
	RtlFillMemory(&infoConnection, sizeof(TDI_CONNECTION_INFORMATION), 0);
	infoConnection.RemoteAddressLength = sizeof(TA_IP_ADDRESS);
	infoConnection.RemoteAddress = &addressIp;

	//TDI_CONNECTION_INFORMATION::RemoteAddress̐ݒ
	RtlFillMemory(&addressIp, sizeof(TA_IP_ADDRESS), 0);
	addressIp.TAAddressCount = 1;
	addressIp.Address[0].AddressLength = TDI_ADDRESS_LENGTH_IP;
	addressIp.Address[0].AddressType = TDI_ADDRESS_TYPE_IP;
	addressIp.Address[0].Address[0].sin_port = portRemote;
	addressIp.Address[0].Address[0].in_addr = addressRemote;

	//IRP̐ݒ
	TdiBuildSendDatagram(pIrp, pDeviceUdp, pFileTransportAddress, NULL, NULL, pMdl, MmGetMdlByteCount(pMdl), &infoConnection);

	//ďo
	status = IoCallDriver(pDeviceUdp, pIrp);
	if (STATUS_PENDING == status)
		KeWaitForSingleObject(&event, Executive, KernelMode, FALSE, NULL);

	status = statusIo.Status;
	if (NULL != pLengthProcessed)
		*pLengthProcessed = statusIo.Information;

	return status;
}

NTSTATUS openTransportAddressUdp(PHANDLE phTransportAddress, PFILE_OBJECT *ppFileTransportAddress)
{
#define LENGTH_FULL_EA_INFORMATION (sizeof(FILE_FULL_EA_INFORMATION)-1 + TDI_TRANSPORT_ADDRESS_LENGTH+1 + sizeof(TA_IP_ADDRESS))
	NTSTATUS status;
	HANDLE hTransportAddress;
	PFILE_OBJECT pFileTransportAddress;
	UNICODE_STRING nameUdp;
	OBJECT_ATTRIBUTES attributeUdp;
	PFILE_FULL_EA_INFORMATION pInformationUdp;
	PTA_IP_ADDRESS pAddressIp;
	IO_STATUS_BLOCK statusIoForUdp;
	UCHAR aBuffer[LENGTH_FULL_EA_INFORMATION];


	//̌
	if (NULL==phTransportAddress || NULL==ppFileTransportAddress)
		return STATUS_INVALID_PARAMETER;

	//OBJECT_ATTRIBUTES̐ݒ
	RtlInitUnicodeString(&nameUdp, L"\\Device\\Udp");
	InitializeObjectAttributes(&attributeUdp, &nameUdp, OBJ_CASE_INSENSITIVE| OBJ_KERNEL_HANDLE, NULL, NULL);

	//FILE_FULL_EA_INFORMATION̐ݒ
	pInformationUdp = PFILE_FULL_EA_INFORMATION(&aBuffer);
	RtlFillMemory(pInformationUdp, LENGTH_FULL_EA_INFORMATION, 0);
	pInformationUdp->NextEntryOffset = 0;
	pInformationUdp->Flags = 0;
	pInformationUdp->EaNameLength = TDI_TRANSPORT_ADDRESS_LENGTH;
	pInformationUdp->EaValueLength = sizeof(TA_IP_ADDRESS);
	RtlCopyMemory(&pInformationUdp->EaName, TdiTransportAddress, TDI_TRANSPORT_ADDRESS_LENGTH);

	//FILE_FULL_EA_INFORMATION::EaName - TA_IP_ADDRESS
	pAddressIp = PTA_IP_ADDRESS(&pInformationUdp->EaName + TDI_TRANSPORT_ADDRESS_LENGTH + 1);
	pAddressIp->TAAddressCount = 1;
	pAddressIp->Address[0].AddressLength = TDI_ADDRESS_LENGTH_IP;
	pAddressIp->Address[0].AddressType = TDI_ADDRESS_TYPE_IP;

	//\Device\UdpJ
	status = ZwCreateFile(&hTransportAddress, FILE_READ_EA| FILE_WRITE_EA| SYNCHRONIZE, &attributeUdp, &statusIoForUdp, NULL, FILE_ATTRIBUTE_NORMAL, 0, FILE_OPEN_IF, 0, pInformationUdp, LENGTH_FULL_EA_INFORMATION);
	if (!NT_SUCCESS(status))
		return status;

	//FILE_OBJECT𓾂
	status = ObReferenceObjectByHandle(hTransportAddress, FILE_ANY_ACCESS, NULL, KernelMode, (PVOID *)&pFileTransportAddress, NULL);
	if (!NT_SUCCESS(status))
	{
		ZwClose(hTransportAddress);

		return status;
	}

	*phTransportAddress = hTransportAddress;
	*ppFileTransportAddress = pFileTransportAddress;

	return status;
}

NTSTATUS closeTransportAddressUdp(HANDLE hTransportAddress, PFILE_OBJECT pFileTransportAddress)
{
	ObDereferenceObject(pFileTransportAddress);
	ZwClose(hTransportAddress);

	return STATUS_SUCCESS;
}
